package dev.fyloz.colorrecipesexplorer.rest

import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeEditUsers
import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeRemoveUsers
import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeViewUsers
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.service.EmployeeGroupService
import dev.fyloz.colorrecipesexplorer.service.EmployeeService
import org.springframework.http.MediaType
import org.springframework.security.access.prepost.PreAuthorize
import org.springframework.web.bind.annotation.*
import java.security.Principal
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse
import javax.validation.Valid

private const val EMPLOYEE_CONTROLLER_PATH = "api/employee"
private const val EMPLOYEE_GROUP_CONTROLLER_PATH = "api/employee/group"

@RestController
@RequestMapping(EMPLOYEE_CONTROLLER_PATH)
class EmployeeController(private val employeeService: EmployeeService) {
    @GetMapping
    @PreAuthorizeViewUsers
    fun getAll() =
        ok(employeeService.getAllForOutput())

    @GetMapping("{id}")
    @PreAuthorizeViewUsers
    fun getById(@PathVariable id: Long) =
        ok(employeeService.getByIdForOutput(id))

    @GetMapping("current")
    fun getCurrent(loggedInEmployee: Principal?) =
        if (loggedInEmployee != null)
            ok(
                with(employeeService) {
                    getById(
                        loggedInEmployee.name.toLong(),
                        ignoreDefaultGroupUsers = false,
                        ignoreSystemUsers = false
                    ).toOutput()
                }
            )
        else
            forbidden()

    @PostMapping
    @PreAuthorizeEditUsers
    fun save(@Valid @RequestBody employee: EmployeeSaveDto) =
        created<EmployeeOutputDto>(EMPLOYEE_CONTROLLER_PATH) {
            with(employeeService) {
                save(employee).toOutput()
            }
        }

    @PutMapping
    @PreAuthorizeEditUsers
    fun update(@Valid @RequestBody employee: EmployeeUpdateDto) =
        noContent {
            employeeService.update(employee)
        }

    @PutMapping("{id}/password", consumes = [MediaType.TEXT_PLAIN_VALUE])
    @PreAuthorizeEditUsers
    fun updatePassword(@PathVariable id: Long, @RequestBody password: String) =
        noContent {
            employeeService.updatePassword(id, password)
        }

    @PutMapping("{employeeId}/permissions/{permission}")
    @PreAuthorizeEditUsers
    fun addPermission(
        @PathVariable employeeId: Long,
        @PathVariable permission: EmployeePermission
    ) = noContent {
        employeeService.addPermission(employeeId, permission)
    }

    @DeleteMapping("{employeeId}/permissions/{permission}")
    @PreAuthorizeEditUsers
    fun removePermission(
        @PathVariable employeeId: Long,
        @PathVariable permission: EmployeePermission
    ) = noContent {
        employeeService.removePermission(employeeId, permission)
    }

    @DeleteMapping("{id}")
    @PreAuthorizeRemoveUsers
    fun deleteById(@PathVariable id: Long) =
        employeeService.deleteById(id)
}

@RestController
@RequestMapping(EMPLOYEE_GROUP_CONTROLLER_PATH)
class GroupsController(
    private val groupService: EmployeeGroupService,
    private val employeeService: EmployeeService
) {
    @GetMapping
    @PreAuthorize("hasAnyAuthority('VIEW_RECIPES', 'VIEW_USERS')")
    fun getAll() =
        ok(groupService.getAllForOutput())

    @GetMapping("{id}")
    @PreAuthorizeViewUsers
    fun getById(@PathVariable id: Long) =
        ok(groupService.getByIdForOutput(id))

    @GetMapping("{id}/employees")
    @PreAuthorizeViewUsers
    fun getEmployeesForGroup(@PathVariable id: Long) =
        ok(with(employeeService) {
            groupService.getEmployeesForGroup(id)
                .map { it.toOutput() }
        })

    @PostMapping("default/{groupId}")
    @PreAuthorizeViewUsers
    fun setDefaultGroup(@PathVariable groupId: Long, response: HttpServletResponse) =
        noContent {
            groupService.setResponseDefaultGroup(groupId, response)
        }

    @GetMapping("default")
    @PreAuthorizeViewUsers
    fun getRequestDefaultGroup(request: HttpServletRequest) =
        ok(with(groupService) {
            getRequestDefaultGroup(request).toOutput()
        })

    @PostMapping
    @PreAuthorizeEditUsers
    fun save(@Valid @RequestBody group: EmployeeGroupSaveDto) =
        created<EmployeeGroupOutputDto>(EMPLOYEE_GROUP_CONTROLLER_PATH) {
            with(groupService) {
                save(group).toOutput()
            }
        }

    @PutMapping
    @PreAuthorizeEditUsers
    fun update(@Valid @RequestBody group: EmployeeGroupUpdateDto) =
        noContent {
            groupService.update(group)
        }

    @DeleteMapping("{id}")
    @PreAuthorizeRemoveUsers
    fun deleteById(@PathVariable id: Long) =
        noContent {
            groupService.deleteById(id)
        }
}

@RestController
@RequestMapping("api")
class LogoutController(private val employeeService: EmployeeService) {
    @GetMapping("logout")
    fun logout(request: HttpServletRequest) =
        ok<Void> {
            employeeService.logout(request)
        }
}
