package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.exception.RestException
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.repository.RecipeStepRepository
import dev.fyloz.colorrecipesexplorer.utils.findDuplicated
import dev.fyloz.colorrecipesexplorer.utils.hasGaps
import org.springframework.http.HttpStatus
import org.springframework.stereotype.Service

interface RecipeStepService : ModelService<RecipeStep, RecipeStepRepository> {
    /** Validates the steps of the given [groupInformation], according to the criteria of [validateSteps]. */
    fun validateGroupInformationSteps(groupInformation: RecipeGroupInformation)

    /**
     *  Validates if the given [steps]. To be valid, the position of each step must be greater or equals to 1 and unique in the set.
     *  There must also be no gap between the positions.
     *  If any of those criteria are not met, an [InvalidGroupStepsPositionsException] will be thrown.
     */
    fun validateSteps(steps: Set<RecipeStep>)
}

@Service
class RecipeStepServiceImpl(recipeStepRepository: RecipeStepRepository) :
  AbstractModelService<RecipeStep, RecipeStepRepository>(recipeStepRepository),
  RecipeStepService {
    override fun idNotFoundException(id: Long) = recipeStepIdNotFoundException(id)
    override fun idAlreadyExistsException(id: Long) = recipeStepIdAlreadyExistsException(id)

    override fun validateGroupInformationSteps(groupInformation: RecipeGroupInformation) {
        if (groupInformation.steps == null) return

        try {
            validateSteps(groupInformation.steps!!)
        } catch (validationException: InvalidStepsPositionsException) {
            throw InvalidGroupStepsPositionsException(groupInformation.group, validationException)
        }
    }

    override fun validateSteps(steps: Set<RecipeStep>) {
        if (steps.isEmpty()) return

        val sortedSteps = steps.sortedBy { it.position }
        val errors = mutableSetOf<InvalidStepsPositionsError>()

        // Check if the first step position is 1
        fun isFirstStepPositionInvalid() =
          sortedSteps[0].position != 1

        // Check if any position is duplicated
        fun getDuplicatedPositionsErrors() =
          sortedSteps
            .findDuplicated { it.position }
            .map { duplicatedStepsPositions(it) }

        // Find all errors and throw if there is any
        if (isFirstStepPositionInvalid()) errors += invalidFirstStepPosition(sortedSteps[0])
        errors += getDuplicatedPositionsErrors()
        if (errors.isEmpty() && steps.hasGaps { it.position }) errors += gapBetweenStepsPositions()
        if (errors.isNotEmpty()) {
            throw InvalidStepsPositionsException(errors)
        }
    }
}

data class InvalidStepsPositionsError(
  val type: String,
  val details: String
)

class InvalidStepsPositionsException(
  val errors: Set<InvalidStepsPositionsError>
) : RestException(
  "invalid-recipestep-position",
  "Invalid steps positions",
  HttpStatus.BAD_REQUEST,
  "The position of steps are invalid",
  mapOf(
    "invalidSteps" to errors
  )
)

class InvalidGroupStepsPositionsException(
  val group: EmployeeGroup,
  val exception: InvalidStepsPositionsException
) : RestException(
  "invalid-groupinformation-recipestep-position",
  "Invalid steps positions",
  HttpStatus.BAD_REQUEST,
  "The position of steps for the group ${group.name} are invalid",
  mapOf(
    "group" to group.name,
    "groupId" to group.id!!,
    "invalidSteps" to exception.errors
  )
) {
    val errors: Set<InvalidStepsPositionsError>
        get() = exception.errors
}

const val INVALID_FIRST_STEP_POSITION_ERROR_CODE = "first"
const val DUPLICATED_STEPS_POSITIONS_ERROR_CODE = "duplicated"
const val GAP_BETWEEN_STEPS_POSITIONS_ERROR_CODE = "gap"

private fun invalidFirstStepPosition(step: RecipeStep) =
  InvalidStepsPositionsError(
    INVALID_FIRST_STEP_POSITION_ERROR_CODE,
    "The position ${step.position} is under the minimum of 1"
  )

private fun duplicatedStepsPositions(position: Int) =
  InvalidStepsPositionsError(
    DUPLICATED_STEPS_POSITIONS_ERROR_CODE,
    "The position $position is duplicated"
  )

private fun gapBetweenStepsPositions() =
  InvalidStepsPositionsError(
    GAP_BETWEEN_STEPS_POSITIONS_ERROR_CODE,
    "There is a gap between steps positions"
  )
